'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _extends = _interopDefault(require('@babel/runtime/helpers/extends'));
var _objectWithoutPropertiesLoose = _interopDefault(require('@babel/runtime/helpers/objectWithoutPropertiesLoose'));
var THREE = require('three');
var React = require('react');
var React__default = _interopDefault(React);
var ResizeObserver = _interopDefault(require('resize-observer-polyfill'));
var _construct = _interopDefault(require('@babel/runtime/helpers/construct'));
var Reconciler = _interopDefault(require('react-reconciler'));
var scheduler = require('scheduler');

function _toPropertyKey(arg) { var key = _toPrimitive(arg, "string"); return typeof key === "symbol" ? key : String(key); }

function _toPrimitive(input, hint) { if (typeof input !== "object" || input === null) return input; var prim = input[Symbol.toPrimitive]; if (prim !== undefined) { var res = prim.call(input, hint || "default"); if (typeof res !== "object") return res; throw new TypeError("@@toPrimitive must return a primitive value."); } return (hint === "string" ? String : Number)(input); }
var roots = new Map();
var emptyObject = {};
var catalogue = {};
var apply = function apply(objects) {
  return catalogue = _extends({}, catalogue, objects);
};
function applyProps(instance, newProps, oldProps, interpolateArray, container) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  if (interpolateArray === void 0) {
    interpolateArray = false;
  }

  if (instance.obj) instance = instance.obj; // Filter equals, events and reserved props
  //console.log(newProps, oldProps)

  var sameProps = Object.keys(newProps).filter(function (key) {
    return newProps[key] === oldProps[key];
  });
  var handlers = Object.keys(newProps).filter(function (key) {
    return typeof newProps[key] === 'function' && key.startsWith('on');
  });
  var filteredProps = [].concat(sameProps, ['children', 'key', 'ref']).reduce(function (acc, prop) {
    var _ = acc[prop],
        rest = _objectWithoutPropertiesLoose(acc, [prop].map(_toPropertyKey));

    return rest;
  }, newProps);

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach(function (_ref) {
      var key = _ref[0],
          value = _ref[1];

      if (!handlers.includes(key)) {
        var root = instance;
        var target = root[key];

        if (key.includes('-')) {
          var entries = key.split('-');
          target = entries.reduce(function (acc, key) {
            return acc[key];
          }, instance);

          if (target && !target.set) {
            // The target is atomic, this forces us to switch the root
            var _entries$reverse = entries.reverse(),
                name = _entries$reverse[0],
                reverseEntries = _entries$reverse.slice(1);

            root = reverseEntries.reverse().reduce(function (acc, key) {
              return acc[key];
            }, instance);
            key = name;
          }
        }

        if (target && target.set) {
          var _target;

          if (target.constructor.name === value.constructor.name) target.copy(value);else if (Array.isArray(value)) (_target = target).set.apply(_target, value);else target.set(value);
        } else root[key] = value;
      }
    }); // Prep interaction handlers

    if (handlers.length) {
      // Add interactive object to central container
      if (container && instance.raycast) container.__interaction.push(instance);
      instance.__handlers = handlers.reduce(function (acc, key) {
        var _extends2;

        return _extends({}, acc, (_extends2 = {}, _extends2[key.charAt(2).toLowerCase() + key.substr(3)] = newProps[key], _extends2));
      }, {});
    } // Call the update lifecycle, if present


    if (instance.__handlers && instance.__handlers.update) instance.__handlers.update(instance);
  }
}

function createInstance(type, _ref2, container) {
  var _ref2$args = _ref2.args,
      args = _ref2$args === void 0 ? [] : _ref2$args,
      props = _objectWithoutPropertiesLoose(_ref2, ["args"]);

  var name = "" + type[0].toUpperCase() + type.slice(1);
  var instance;
  if (type === 'primitive') instance = props.object;else {
    var target = catalogue[name] || THREE[name];
    instance = Array.isArray(args) ? _construct(target, args) : new target(args);
  }
  applyProps(instance, props, {}, false, container);
  return !instance.isObject3D ? {
    obj: instance,
    parent: undefined
  } : instance;
}

function appendChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.add(child);else {
      child.parent = parentInstance; // The name attribute implies that the object attaches itself on the parent

      if (child.obj.name) {
        if (parentInstance.obj) parentInstance = parentInstance.obj;
        var target = parentInstance[child.obj.name];
        if (Array.isArray(target)) target.push(child.obj);else parentInstance[child.obj.name] = child.obj;
      }
    }
  }
}

function removeChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) {
      parentInstance.remove(child);
      if (child.dispose) child.dispose();
    } else {
      child.parent = undefined;

      if (child.obj.name) {
        if (parentInstance.obj) parentInstance = parentInstance.obj; // Remove attachment

        var target = parentInstance[child.obj.name];
        if (Array.isArray(target)) parentInstance[child.obj.name] = target.filter(function (x) {
          return x !== child.obj;
        });else parentInstance[child.obj.name] = undefined;
      }

      if (child.obj.dispose) child.obj.dispose();
    }
  }
}

function insertBefore(parentInstance, child, beforeChild) {
  if (child) {
    if (child.isObject3D) {
      child.parent = parentInstance;
      child.dispatchEvent({
        type: 'added'
      }); // TODO: the order is out of whack if data objects are present, has to be recalculated

      var index = parentInstance.children.indexOf(beforeChild);
      parentInstance.children = [].concat(parentInstance.children.slice(0, index), [child], parentInstance.children.slice(index));
    } else child.parent = parentInstance;
  }
}

var Renderer = Reconciler({
  now: scheduler.unstable_now,
  createInstance: createInstance,
  removeChild: removeChild,
  appendChild: appendChild,
  insertBefore: insertBefore,
  supportsMutation: true,
  isPrimaryRenderer: false,
  schedulePassiveEffects: scheduler.unstable_scheduleCallback,
  cancelPassiveEffects: scheduler.unstable_cancelCallback,
  appendInitialChild: appendChild,
  appendChildToContainer: appendChild,
  removeChildFromContainer: removeChild,
  insertInContainerBefore: insertBefore,
  commitUpdate: function commitUpdate(instance, updatePayload, type, oldProps, newProps, fiber) {
    if (instance.isObject3D) {
      applyProps(instance, newProps, oldProps);
    } else {
      // This is a data object, let's extract critical information about it
      var parent = instance.parent;

      var _newProps$args = newProps.args,
          argsNew = _newProps$args === void 0 ? [] : _newProps$args,
          restNew = _objectWithoutPropertiesLoose(newProps, ["args"]);

      var _oldProps$args = oldProps.args,
          argsOld = _oldProps$args === void 0 ? [] : _oldProps$args,
          restOld = _objectWithoutPropertiesLoose(oldProps, ["args"]); // If it has new props or arguments, then it needs to be re-instanciated


      if (argsNew.some(function (value, index) {
        return value !== argsOld[index];
      })) {
        // Next we create a new instance and append it again
        var newInstance = createInstance(type, newProps);
        removeChild(parent, instance);
        appendChild(parent, newInstance); // Switch instance

        instance.obj = newInstance.obj;
        instance.parent = newInstance.parent;
      } else {
        // Otherwise just overwrite props
        applyProps(instance, restNew, restOld);
      }
    }
  },
  getPublicInstance: function getPublicInstance(instance) {
    return instance;
  },
  getRootHostContext: function getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },
  getChildHostContext: function getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },
  createTextInstance: function createTextInstance() {},
  finalizeInitialChildren: function finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },
  prepareUpdate: function prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },
  shouldDeprioritizeSubtree: function shouldDeprioritizeSubtree(type, props) {
    return false;
  },
  prepareForCommit: function prepareForCommit() {},
  resetAfterCommit: function resetAfterCommit() {},
  shouldSetTextContent: function shouldSetTextContent(props) {
    return false;
  }
});
function render(element, container) {
  var root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  var root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, function () {
    return roots.delete(container);
  });
}

var stateContext = React__default.createContext();

function useMeasure() {
  var ref = React.useRef();

  var _useState = React.useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
      bounds = _useState[0],
      set = _useState[1];

  var _useState2 = React.useState(function () {
    return new ResizeObserver(function (_ref) {
      var entry = _ref[0];
      return set(entry.contentRect);
    });
  }),
      ro = _useState2[0];

  React.useEffect(function () {
    if (ref.current) ro.observe(ref.current);
    return function () {
      return ro.disconnect();
    };
  }, []);
  return [{
    ref: ref
  }, bounds];
}

var Canvas = React__default.memo(function (_ref2) {
  var children = _ref2.children,
      props = _ref2.props,
      camera = _ref2.camera,
      style = _ref2.style,
      rest = _objectWithoutPropertiesLoose(_ref2, ["children", "props", "camera", "style"]);

  // Local, reactive state
  var canvas = React.useRef();

  var _useState3 = React.useState(false),
      ready = _useState3[0],
      setReady = _useState3[1];

  var _useMeasure = useMeasure(),
      bind = _useMeasure[0],
      size = _useMeasure[1];

  var _useState4 = React.useState([]),
      intersects = _useState4[0],
      setIntersects = _useState4[1];

  var _useState5 = React.useState('default'),
      cursor = _useState5[0],
      setCursor = _useState5[1];

  var _useState6 = React.useState(function () {
    return new THREE.Raycaster();
  }),
      raycaster = _useState6[0];

  var _useState7 = React.useState(function () {
    return new THREE.Vector2();
  }),
      mouse = _useState7[0];

  var _useState8 = React.useState(function () {
    var cam = new THREE.PerspectiveCamera(75, 0, 0.1, 1000);
    cam.position.z = 5;
    if (camera) applyProps(cam, camera, {});
    return cam;
  }),
      defaultCam = _useState8[0],
      _setDefaultCamera = _useState8[1]; // Public state


  var state = React.useRef({
    ready: false,
    subscribers: [],
    manual: false,
    active: true,
    canvas: undefined,
    gl: undefined,
    camera: undefined,
    scene: undefined,
    size: undefined,
    viewport: function viewport(target) {
      if (target === void 0) {
        target = new THREE.Vector3(0, 0, 0);
      }

      var distance = state.current.camera.position.distanceTo(target);
      var fov = THREE.Math.degToRad(state.current.camera.fov); // convert vertical fov to radians

      var height = 2 * Math.tan(fov / 2) * distance; // visible height

      var width = height * state.current.camera.aspect;
      return {
        width: width,
        height: height
      };
    },
    subscribe: function subscribe(fn, main) {
      state.current.subscribers.push(fn);
      return function () {
        return state.current.subscribers = state.current.subscribers.filter(function (s) {
          return s === fn;
        });
      };
    },
    setManual: function setManual(takeOverRenderloop) {
      state.current.manual = takeOverRenderloop;

      if (takeOverRenderloop) {
        // In manual mode items shouldn't really be part of the internal scene which has adverse effects
        // on the camera being unable to update without explicit calls to updateMatrixWorl()
        state.current.scene.children.forEach(function (child) {
          return state.current.scene.remove(child);
        });
      }
    },
    setDefaultCamera: function setDefaultCamera(cam) {
      state.current.camera = cam;

      _setDefaultCamera(cam);
    }
  }); // Writes locals into public state for distribution among subscribers, context, etc

  React.useEffect(function () {
    state.current.ready = ready;
    state.current.size = size;
    state.current.aspect = size.width / size.height;
    state.current.camera = defaultCam;
  }, [ready, size, defaultCam]); // Component mount effect, creates the webGL render context

  React.useEffect(function () {
    state.current.gl = new THREE.WebGLRenderer(_extends({
      canvas: canvas.current,
      antialias: true,
      alpha: true
    }, props));
    state.current.gl.setPixelRatio((window.devicePixelRatio || 2) * 0.5);
    state.current.gl.setClearAlpha(0);
    state.current.canvas = canvas.current;
    state.current.scene = new THREE.Scene();
    state.current.scene.__interaction = [];

    var renderLoop = function renderLoop() {
      if (!state.current.active) return;
      requestAnimationFrame(renderLoop);

      if (state.current.ready) {
        state.current.subscribers.forEach(function (fn) {
          return fn(state.current);
        });

        if (!state.current.manual && state.current.scene.children.length) {
          state.current.gl.render(state.current.scene, state.current.camera);
        }
      }
    }; // Start render-loop


    requestAnimationFrame(renderLoop); // Clean-up

    return function () {
      state.current.active = false;
      unmountComponentAtNode(state.current.scene); // TODO: Clean up and dispose scene ...
    };
  }, []); // Adjusts default camera

  React.useEffect(function () {
    if (ready) {
      state.current.gl.setSize(size.width, size.height);
      state.current.camera.aspect = state.current.aspect;
      state.current.camera.radius = (size.width + size.height) / 4;
      state.current.camera.updateProjectionMatrix();
    }
  }, [ready, size, defaultCam]);
  var intersect = React.useCallback(function (event, fn) {
    var canvasRect = canvas.current.getBoundingClientRect();
    var x = (event.clientX - canvasRect.left) / (canvasRect.right - canvasRect.left) * 2 - 1;
    var y = -((event.clientY - canvasRect.top) / (canvasRect.bottom - canvasRect.top)) * 2 + 1;
    mouse.set(x, y, 0.5);
    raycaster.setFromCamera(mouse, state.current.camera); // TODO only inspect onbjects that have handlers

    var hits = raycaster.intersectObjects(state.current.scene.__interaction, true).filter(function (h) {
      return h.object.__handlers;
    });
    hits.forEach(fn);
    return hits;
  }, []);
  React.useEffect(function () {
    var hovered = {};

    var handleMove = function handleMove(event) {
      var hover = false;
      var hits = intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(function (object) {
        if (!hits.length || !hits.find(function (i) {
          return i.object === object;
        })) {
          if (object.__handlers.unhover) {
            object.__handlers.unhover();
          }

          delete hovered[object.uuid];
        }
      });
      /*if (intersects.length !== hits.length || intersects.some((h, index) => h.object !== hits[index].object))
        setIntersects(hits)*/
    };

    var handleClick = function handleClick(event) {
      var clicked = {};
      intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.click && !clicked[object.uuid]) {
          clicked[object.uuid] = object;
          handlers.click(data);
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    window.addEventListener('mouseup', handleClick, {
      passive: true
    });
    return function () {
      window.removeEventListener('mousemove', handleMove);
      window.removeEventListener('mouseup', handleClick);
    };
  }); // This component is a bridge into the three render context, when it gets rendererd
  // we know we are ready to compile shaders, call subscribers, etc

  var IsReady = React.useCallback(function () {
    React.useEffect(function () {
      return void setReady(true);
    }, []);
    return null;
  }, []); // Render v-dom into scene

  React.useEffect(function () {
    if (size.width > 0 && size.height > 0) {
      render(React__default.createElement(stateContext.Provider, {
        value: _extends({}, state.current)
      }, React__default.createElement(IsReady, null), children), state.current.scene);
    }
  }); // Render the canvas into the dom

  return React__default.createElement("div", _extends({}, bind, rest, {
    style: _extends({
      cursor: cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }, style)
  }), React__default.createElement("canvas", {
    ref: canvas
  }));
});

function useRender(fn, takeOverRenderloop) {
  var _useContext = React.useContext(stateContext),
      subscribe = _useContext.subscribe,
      setManual = _useContext.setManual; // This calls into the host to inform it whether the render-loop is manual or not


  React.useMemo(function () {
    return takeOverRenderloop && setManual(true);
  }, [takeOverRenderloop]);
  React.useEffect(function () {
    // Subscribe to the render-loop
    var unsubscribe = subscribe(fn, takeOverRenderloop);
    return function () {
      // Call subscription off on unmount
      unsubscribe();
      if (takeOverRenderloop) setManual(false);
    };
  }, []);
}
function useThree(fn) {
  var _useContext2 = React.useContext(stateContext),
      subscribe = _useContext2.subscribe,
      props = _objectWithoutPropertiesLoose(_useContext2, ["subscribe"]);

  return props;
}
function useCamera(instance) {
  for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  var _useThree = useThree(),
      size = _useThree.size,
      setDefaultCamera = _useThree.setDefaultCamera;

  var _useState = React.useState(function () {
    var camera = _construct(instance, args);

    setDefaultCamera(camera);
    return camera;
  }),
      camera = _useState[0];

  React.useMemo(function () {
    camera.aspect = size.width / size.height;
    camera.radius = (size.width + size.height) / 4;
    camera.updateProjectionMatrix();
  }, [size]);
  return camera;
} // TODO

function useSelection() {}

exports.Canvas = Canvas;
exports.render = render;
exports.unmountComponentAtNode = unmountComponentAtNode;
exports.apply = apply;
exports.applyProps = applyProps;
exports.useRender = useRender;
exports.useThree = useThree;
exports.useCamera = useCamera;
exports.useSelection = useSelection;
